// Get 1-wire Library here: http://www.pjrc.com/teensy/td_libs_OneWire.html
#include <OneWire.h>
#include <PID_v1.h>
#include <avr/wdt.h>
//Get DallasTemperature Library here:  http://milesburton.com/Main_Page?title=Dallas_Temperature_Control_Library
#include <DallasTemperature.h>
// Wire (I2C) Library
#include <Wire.h>
#include <Adafruit_ADS1015.h>
Adafruit_ADS1115 ads(0x48);

/*-----( Declare Constants and Pin Numbers )-----*/
// Data wire is plugged into port 8 on the Arduino (can be changed)
#define ONE_WIRE_BUS 7    // NOTE: No ";" on #define


// Setup a oneWire instance to communicate with any OneWire devices
// (not just Maxim/Dallas temperature ICs)
OneWire oneWire(ONE_WIRE_BUS);

// Pass address of our oneWire instance to Dallas Temperature.
DallasTemperature sensors(&oneWire);
#include <LiquidCrystal_I2C.h>

// Start the LCD display library
LiquidCrystal_I2C lcd(0x27, 2, 1, 0, 4, 5, 6, 7, 3, POSITIVE);  // Set the LCD I2C address 0x3F or 0x27

//===========================MAIN CHARGER SETTINGS, REFER TO YOUR BATTERY MANUAL==================================================================

float batv            = 48;     // nominal battery voltage 24 OR 48, could work with 12V with 18v power supply for FET driver chips on pcb
double floatVolts        = 54.0;   // User setting for the FLOAT voltag
double absorbCharge      = 59.2;   // User setting for the ABSORPTION voltage
long timedOut         = 200;    // ABSORPTION MINUTES
int startAllOverAgain = 51.5;     // User setting for resetting the charge cycle, usually overnight or after heavy drawdown
int startOverTime     = 5;   //?? Minutes below Start over voltage to take effect
int Ventfanvolts      = 51.0;   // Turn on battery vent fan


//===========================ADVANCED SETTINGS===================================================================================================


float R1 = 105000.0;            // resistance of R1 (100K)
float R2 = 8000.0;              // resistance of R2 (8K2) For setting up voltage divider to get correct voltage reading Use in conjunction with trim pot on pcb
int pwmmin = 12;                 // Minimum Output level 0 = will go as low as no output, set higher WITH CAUTION if you want to stop a GTI from restarting during the day if unsure leave at zero
int SOTimer = 0;
const int numReadings = 10;
int readings;                   // the readings from the analog input
int readIndex = 0;              // the index of the current reading
int total = 0;                  // the running total
int average = 0;                // the average
double Setpoint = 0;
//int voltsinputPin = 3;
double battemp;
double tcomp;
double tcomp1;
double volts = 0.0;
double vin = 0.0;
int value = 0;
long fullTime = 0;            // timer for float stage
int PWMout1 = 5;              // fet driver connected to this pin
int PWMout2 = 6;              // fet driver connected to this pin
int PWMHW = 10;               //PWM Output to Hot Water Heater SSR
int PWMLOAD = 11;             //PWM Output to Load SSR
double pulseVal = 0;             // initilise the pulse width to zero of 255
double hwval = 0;
long absorbTime = 0;            // initialise the time to zero for the absorb
int weAreHereNow = 0;         // initialize the start stage as zero
int weWhereHere = 0;

unsigned long start, finished, elapsed;
unsigned long astart, afinished, aelapsed, apause, fstart;
boolean r = false;
const int buttonFloat = 9;     // the number of the pushbutton pin
const int buttonAbsorb = 8;     // the number of the pushbutton pin
int buttonFState = 0;
int buttonAState = 0;

double Kp = 55, Ki = 24, Kd = 1;
PID myPID(&vin, &pulseVal, &Setpoint, Kp, Ki, Kd, DIRECT);


void setup() {
  Serial.begin(9600);
  sensors.begin();
  ads.begin();
  // initialize all the readings to 0:

  { lcd.begin(20, 4);
    pinMode(PWMout1, OUTPUT);
    pinMode(PWMout2, OUTPUT);      // sets the pin 6 as an output for the fet drive
    pinMode(PWMHW, OUTPUT);
    pinMode(PWMLOAD, OUTPUT);
    pinMode(A0, OUTPUT);           //Vent Fan Output
    pinMode(A1, OUTPUT);
  }
  myPID.SetMode(AUTOMATIC);
  wdt_enable(WDTO_4S); //Enable watchdog to reset if not restarted within 4 seconds
}
void loop() {
  wdt_reset(); //Resets Watchdog If the NANO hangs this will cause a reset if the timer is not restarted
  int16_t adc0, adc3;
  adc0 = ads.readADC_SingleEnded(0);
  adc3 = ads.readADC_SingleEnded(3);
  {
    // subtract the last reading:
    total = total - readings;
    // read from the sensor:
    readings = adc3;
    // add the reading to the total:
    total = total + readings;
    // advance to the next position in the array:
    readIndex = readIndex + 1;

    // if we're at the end of the array...
    if (readIndex >= numReadings) {
      // ...wrap around to the beginning:
      readIndex = 0;
    }

    // calculate the average:
    average = total / numReadings;

    volts = average;
    vin = volts / 38;    // Voltage divider calculation
  }
  buttonFState = digitalRead(buttonFloat);
  buttonAState = digitalRead(buttonAbsorb);

  if (buttonFState == HIGH) {
    weAreHereNow = 2;
  }// Enter float mode:

  if (buttonAState == HIGH) {
    weAreHereNow = 1;
  }// Enter Absorb mode:

  sensors.requestTemperatures(); // Send the command to get temperature readings

  battemp = (sensors.getTempCByIndex(0)); // convert sensor value to degrees celcius

  if (battemp == -127.00) // Measurement failed or no device found, if no temp sensor temperature comp has no effect
  {
    battemp = tcomp1;    // prevents error if sensor fails to return a result occasionally
  }
  else
  {
    tcomp = (25 - battemp) * (batv * 0.0025);   //used to calculate voltage set point with temperature compensation applied
    tcomp1 = battemp;
  }

  lcd.setCursor(0, 0);
  lcd.print("Batt V = ");
  //Serial.print ("Batt V = ");
  lcd.setCursor(9, 0);
  lcd.print(vin);            //Actual battery volts
  //Serial.println (vin);
  lcd.setCursor(15, 0);
  lcd.print("C");
  lcd.setCursor(16, 0);
  lcd.print(tcomp);         //Temperature compensation voltage adjustment value add or subtract for difference from 25 degrees
  lcd.setCursor(0, 2);
  lcd.print("P%   ");
  lcd.setCursor(2, 2);
  lcd.print(pulseVal / 2.55, 0); // print the PWM out as a percentage from 0 - 100 (100% = charging at full avaiable power)lcd.print(fanpwmH / 2.55, 0);
 // Serial.print ("PWM % =  ");
  //Serial.println(pulseVal / 2.55); // print the PWM out as a percentage from 0 - 100 (100% = charging at full avaiable power)
  lcd.setCursor(9, 2);
  lcd.print("HW%   ");
  lcd.setCursor(11, 2);
  lcd.print(hwval / 2.55, 0);
  {

    if (vin < startAllOverAgain )           //charge cylce starts here
    {
      lcd.setCursor(0, 1);
      lcd.print("Pre-bulk volts ");
      start = millis();                   //sets start time for when bulk charge starts
      Setpoint = 60;
    }


    if (weAreHereNow == 0 && vin > startAllOverAgain)    // if stage 1 then we do this routine
    {
      lcd.setCursor(0, 1);
      lcd.print("Bulk  = ");
      lcd.setCursor (8, 1);
      lcd.print((millis() - start) / 60000); // Displays how many minutes since bulk charge started
      lcd.print(" min");
      Setpoint = 60;
    }

    if (weAreHereNow == 1)    // if stage 2 then we do this screen routine

    {
      lcd.setCursor(0, 1);
      lcd.print("Absorb        ");
      lcd.setCursor (7, 1);
      lcd.print(absorbTime);          // Display minutes actually at absorb voltage
      lcd.setCursor (11, 1);
      lcd.print("< Abs");
      lcd.setCursor (17, 1);
      lcd.print(apause / 60000);      // Display minutes below absorb voltage since absorb started
    }

    if (weAreHereNow == 2)
    {
      lcd.setCursor(0, 1);
      lcd.print("Float = ");
      lcd.setCursor (9, 1);
      lcd.print(fullTime);           // Display minutes since float was started
      lcd.setCursor (10, 1);
      lcd.print(" min ");
      lcd.print (SOTimer);

    }

    lcd.setCursor(0, 3);
    lcd.print("BTemp");
    lcd.setCursor (6, 3);
    lcd.print (battemp);             // Display actual battery temperature
    lcd.setCursor (11, 3);
    lcd.print (Setpoint);
  }

  if (vin >= (absorbCharge + tcomp) && weAreHereNow == 0) // if we achieve 59v (bulk chg) AND we were in stage zero then change weAreHereNow to 1 as a flag ie absorb
  {
    weAreHereNow = 1;               //   weAreHereNow=1 is absorb status
  }

  //=============bulk routine==============


  if ( weAreHereNow == 0)           // if in stage 1 then check voltage against bulk charge constant ... can be set in definitions
  { pulseVal = 255;                 // was not up to bulkCharge, then increase pulse width incrementally
    astart = millis();              //sets start time for when absorb charge starts
    Setpoint = 60;
  }

  //==========absord routine==============

  if ( weAreHereNow == 1 )            // if we are in absorb then do this
  {
    if  (vin >= ((absorbCharge + tcomp) - 0.5)) { // This section is keep track of the time at or just under absorb voltage
      aelapsed = ((millis() - astart) - apause);  // aelapsed = time from the absorb cycle starting less the time below (absorbCharge + tcomp) - 0.5)
      afinished = millis() - apause;              // afinished is used to calculate time below absorb
      absorbTime = (aelapsed) / 60000;            // convert time to minutes

    } else {
      apause = (millis() - afinished);            //time below absorb set point since absorb started
    }
    if (absorbTime >= timedOut)                   // once we reach timedOut, we have finished our absorb sequence, change status flag weAreHereNow to 2, and change the voltage now to float value
    {
      weAreHereNow = 2;
      fstart = millis();
    }

    Setpoint = (absorbCharge + tcomp);

    myPID.Compute();

//    if ( pulseVal <= pwmmin)
//    {
//      pulseVal = pwmmin;
//    }
  }
  //=========float routine=======


  if ( weAreHereNow == 2)      // are we in float charge stage

  {
    if  (vin > ((floatVolts + tcomp) - 0.5))

    {
      fullTime = (millis() - fstart) / 60000;   // increment float timer
    }


    Setpoint = (floatVolts + tcomp);

    myPID.Compute();

//    if ( pulseVal <= pwmmin)
//    {
//      pulseVal = pwmmin;
//    }

  }
  //==============================write the pulse to the fet======================
  analogWrite (PWMout2, pulseVal);
    lcd.setCursor(6, 2);
    lcd.print("   ");
    lcd.setCursor(6, 2);
  lcd.print(pulseVal/2.55, 0);
  if ( pulseVal <= pwmmin)
  {
    pulseVal = pwmmin;
  }
  analogWrite (PWMout1, pulseVal); // actually write the data to the fet via pin 5 ...

  if (vin >= (Setpoint - 0.7))
  {
    hwval = (((vin + 0.7) - (Setpoint)) * 510);
    hwval = constrain(hwval, 0, 255);
  }
  else
  {
    hwval = 0;
  }
  analogWrite (PWMHW, hwval);

  if (vin <= (Ventfanvolts - 0.5))
  {
    digitalWrite(A0, LOW);
    lcd.setCursor(18, 2);
  lcd.print("  ");
  }

  if (vin >= Ventfanvolts)
  {
    digitalWrite(A0, HIGH);
    lcd.setCursor(18, 2);
  lcd.print("VF");
  }

  if (vin <= (Ventfanvolts - 0.5))
  {
    digitalWrite(A0, LOW);
  }

  //=============================================
  //  if (vin < (startAllOverAgain))// && (SOTimer == 0))      // Voltage level to restart charge cycle when voltage drops below threshold due to night or very heavy load
  //{
  //  SOTimer = millis();
  //}
  //if (vin => (startAllOverAgain))
  //{
  //  SOTimer = 0;
  //}
  //  if ((startOverTime * 1000) > (millis() - SOTimer))
  if (vin < (startAllOverAgain + tcomp))        // Voltage level to restart charge cycle when voltage drops below threshold due to night or very heavy load
  {
    weAreHereNow = 0;             //reset values for start of next charge cycle
    absorbTime = 0;
    astart = 0;
    apause = 0;
    aelapsed = 0;
    fstart = 0;
    fullTime = 0;
  }
}
